@@ -8,12 +8,11 @@ from __future__ import (
 )
 
 import logging
-from glob import glob
-from tempfile import TemporaryDirectory
 from typing import Dict
 
 import requests
 from pdf2image import convert_from_bytes, convert_from_path
+from prettytable import PrettyTable
 
 from disclosure_extractor.calculate import estimate_investment_net_worth, color
 from disclosure_extractor.data_processing import process_document
@@ -32,65 +31,15 @@ from disclosure_extractor.judicial_watch_utils import (
 from disclosure_extractor.post_processing import _fine_tune_results
 
 
-def print_results(results):
-    """Display the extraction in a nice table printout."""
-    cd = {}
-    for k, v in results["sections"].items():
-        columns = results["sections"][k]["fields"]
-        cd[k] = []
-        max_lengths = [len(x) for x in columns]
-        if v["rows"] != {}:
-            x = cd[k]
-            for row in v["rows"]:
-                if len(row.items()) != len(columns):
-                    continue
-                r = []
-                lengths = []
-                if max_lengths == None:
-                    max_lengths = [0] * len(row.items())
-                for a, cell in row.items():
-                    if cell["text"] is None:
-                        r.append("")
-                        lengths.append(0)
-                        continue
-                    if cell["is_redacted"] == True:
-                        r.append(cell["text"] + " ⬛⬛⬛ ")
-                        lengths.append(len(cell["text"] + " ⬛⬛⬛ ") + 1)
-                    else:
-                        r.append(cell["text"])
-                        lengths.append(len(cell["text"]) + 1)
-                max_lengths = [
-                    x if x > y else y for x, y in zip(lengths, max_lengths)
-                ]
-                if len("".join(r).strip()) == 0:
-                    continue
-                x.append(r)
-
-            headers = [x.ljust(y) for x, y in zip(columns, max_lengths)]
-            print("\n")
-            print(
-                "|",
-                "-" * len(" | ".join(headers)),
-                "|",
-            )
-            print("\033[1m", end="")
-            print("|", k.ljust(len(" | ".join(headers)) - 1), "\033[0m", "|")
-            print(
-                "|",
-                "-" * len(" | ".join(headers)),
-                "|",
-            )
-            print("\033[1m", end="")
-            print("|", " | ".join(headers), "|", "\033[0m")
-            print("|", "-" * len(" | ".join(headers)), "|")
-            for item in cd[k]:
-                clean_row = [x.ljust(y) for x, y in zip(item, max_lengths)]
-                print("|", " | ".join(clean_row), "|")
-            cd[k] = x
-            print("|", "_" * len(" | ".join(headers)), "|")
+def display_wealth(results: Dict[str, str]) -> None:
+    """Print data in nice neat tables
 
+    :param results: Results of extraction
+    :return: None
+    """
     if "wealth" not in results.keys():
         return
+
     print("\n", "Wealth", "\n===============================")
     net_worth = results["wealth"]["investment_net_worth"]
     gains = results["wealth"]["income_gains"]
@@ -142,6 +91,19 @@ def print_results(results):
     print("\n")
 
 
+def display_table(results: Dict) -> None:
+    """Display information"""
+    for title, sect in results["sections"].items():
+        print(title)
+        t = PrettyTable(sect["fields"])
+        for row in sect["rows"]:
+            r = [row[field]["text"] for field in sect["fields"]]
+            t.add_row(r)
+        print(t, end="\n\n")
+
+    display_wealth(results)
+
+
 def process_financial_document(
     file_path=None, url=None, pdf_bytes=None, show_logs=None, resize_pdf=True
 ):
@@ -195,7 +157,7 @@ def process_financial_document(
 
     logging.info("Extracting content from financial disclosure")
     results = process_document(
-        document_structure, pages, show_logs, resize=resize_pdf
+        document_structure, pages, show_logs, resize_pdf
     )
     results["page_count"] = page_total
     results["pdf_size"] = len(pdf_bytes)
@@ -233,7 +195,7 @@ def process_judicial_watch(
         pdf_bytes = requests.get(url, stream=True).content
 
     # Turn the PDF into an array of images
-    pages = convert_from_bytes(pdf_bytes)
+    pages = convert_from_bytes(pdf_bytes, thread_count=10)
     page_total = len(pages)
     logging.info("Document is %s pages long" % page_total)
 
@@ -285,47 +247,39 @@ def extract_financial_document(
 
     if show_logs:
         logging.getLogger().setLevel(logging.INFO)
-    with TemporaryDirectory() as dir:
-        if pdf_bytes:
-            convert_from_bytes(
-                pdf_bytes, thread_count=4, output_folder=dir, fmt="tiff"
-            )
-        else:
-            convert_from_path(
-                file_path, thread_count=4, output_folder=dir, fmt="tiff"
-            )
-        page_paths = sorted(glob(f"{dir}/*.tif"))
-        logging.info("Document is %s pages long" % len(page_paths))
-        logging.info("Determining document structure.")
-
-        try:
-            document_structure, check_count = extract_contours_from_page(
-                page_paths, resize=resize
-            )
-        except:
-            try:
-                logging.info("Switch resizing again")
-                resize = False if resize else True
-                (
-                    document_structure,
-                    check_count,
-                ) = extract_contours_from_page(page_paths, resize=resize)
-            except:
-                return {"success": False, "msg": CheckboxesNotFound}
-
-        if check_count < 8:
-            logging.warning("Failed to extract document structure")
-            return {
-                "success": False,
-                "msg": "Failed to process document properly",
-                "checkbox_count_found": check_count,
-            }
-
-        logging.info("Extracting content from financial disclosure")
-        results = process_document(
-            document_structure, page_paths, show_logs, resize=resize
+    logging.info("Starting pdf to image conversion")
+    if pdf_bytes:
+        xpages = convert_from_bytes(
+            pdf_bytes, thread_count=10, fmt="jpg", dpi=300
         )
-        results["page_count"] = len(page_paths)
+    else:
+        xpages = convert_from_path(
+            file_path, thread_count=10, fmt="jpg", dpi=300
+        )
+
+    logging.info("Document is %s pages long" % len(xpages))
+    logging.info("Determining document structure.")
+
+    pages = []
+    if resize:
+        for page in xpages:
+            pg = page.resize((1653, 2180))
+            pages.append(pg)
+
+    document_structure = extract_contours_from_page(pages)
+
+    if document_structure["found_count"] < 8:
+        logging.warning("Failed to extract document structure")
+        return {
+            "success": False,
+            "msg": "Failed to process document properly",
+            "checkbox_count_found": document_structure["found_count"],
+        }
+
+    logging.info("Extracting content from financial disclosure")
+    results = process_document(document_structure, pages)
+    results["page_count"] = len(pages)
+
     results["pdf_size"] = ""
     results["wealth"] = estimate_investment_net_worth(results)
     results["success"] = True
@@ -334,7 +288,4 @@ def extract_financial_document(
     # Cleanup raw data & update document structure
     cleaned_data = _fine_tune_results(results)
 
-    if show_logs:
-        print_results(cleaned_data)
-
     return cleaned_data
